
@if not exist "D:\Program Files\Borland\CBuilder4\Bin\tasm32.exe" goto TryBccOnC
@set BCC_TASM="D:\Program Files\Borland\CBuilder4\Bin\tasm32.exe"
@goto Compile

:TryBccOnC
@if not exist "C:\Program Files\Borland\CBuilder4\Bin\tasm32.exe" goto NoBCC
@set BCC_TASM="C:\Program Files\Borland\CBuilder4\Bin\tasm32.exe"
@goto Compile

:Compile
%BCC_TASM% /ml "../atoms/raster_BCC.asm" "../atoms/raster_BCC.obj"
%BCC_TASM% /ml img_tool_BCC.asm
@goto End

:NoBCC
echo "NO BCC4 compiler found"

:End

