// Please do not compile this file into a library, insert it into main project
// instead. Sometimes a global ctor of MMX_SSE_initialiser is not called.
//
// This wrapper is freeware, you can use it whenewer you need.
#include <stdio.h>
#include "raster.h"

#if defined(USE_MMX) || defined(USE_SSE)


extern "C" void MMX_SSE_init(void);


class MMX_SSE_initialiser
{
public:
  MMX_SSE_initialiser(void);
};


MMX_SSE_initialiser::MMX_SSE_initialiser(void)
{
  MMX_SSE_init();
}


MMX_SSE_initialiser MMX_SSE_InitObj;	/// Initialiser ctror should be called before main()


#endif
